@@ -8,12 +8,13 @@ use crate::runtime::vm::memory::{Memory, RuntimeMemoryCreator};
 use crate::runtime::vm::table::{Table, TableElement, TableElementType};
 use crate::runtime::vm::vmcontext::{
     VMBuiltinFunctionsArray, VMContext, VMFuncRef, VMFunctionImport, VMGlobalDefinition,
-    VMGlobalImport, VMMemoryDefinition, VMMemoryImport, VMOpaqueContext, VMRuntimeLimits,
-    VMTableDefinition, VMTableImport,
+    VMGlobalImport, VMMemoryDefinition, VMMemoryImport, VMOpaqueContext, VMStoreContext, VMTable,
+    VMTableDefinition, VMTagDefinition, VMTagImport,
 };
 use crate::runtime::vm::{
-    ExportFunction, ExportGlobal, ExportMemory, ExportTable, GcStore, Imports, ModuleRuntimeInfo,
-    SendSyncPtr, VMFunctionBody, VMGcRef, VMStore, VMStoreRawPtr, VmPtr, VmSafe, WasmFault,
+    ExportFunction, ExportGlobal, ExportMemory, ExportTable, ExportTag, GcStore, Imports,
+    ModuleRuntimeInfo, SendSyncPtr, VMFunctionBody, VMGcRef, VMStore, VMStoreRawPtr, VmPtr, VmSafe,
+    WasmFault,
 };
 use crate::store::{StoreInner, StoreOpaque};
 use crate::{prelude::*, StoreContextMut};
@@ -26,11 +27,13 @@ use core::ptr::NonNull;
 use core::sync::atomic::AtomicU64;
 use core::{mem, ptr};
 use sptr::Strict;
+#[cfg(feature = "gc")]
+use wasmtime_environ::ModuleInternedTypeIndex;
 use wasmtime_environ::{
     packed_option::ReservedValue, DataIndex, DefinedGlobalIndex, DefinedMemoryIndex,
-    DefinedTableIndex, ElemIndex, EntityIndex, EntityRef, EntitySet, FuncIndex, GlobalIndex,
-    HostPtr, MemoryIndex, Module, ModuleInternedTypeIndex, PrimaryMap, PtrSize, TableIndex,
-    TableInitialValue, TableSegmentElements, Trap, VMOffsets, VMSharedTypeIndex, WasmHeapTopType,
+    DefinedTableIndex, DefinedTagIndex, ElemIndex, EntityIndex, EntityRef, EntitySet, FuncIndex,
+    GlobalIndex, HostPtr, MemoryIndex, Module, PrimaryMap, PtrSize, TableIndex, TableInitialValue,
+    TableSegmentElements, TagIndex, Trap, VMOffsets, VMSharedTypeIndex, WasmHeapTopType,
     VMCONTEXT_MAGIC,
 };
 #[cfg(feature = "wmemcheck")]
@@ -168,7 +171,7 @@ impl InstanceAndStore {
 /// A type that roughly corresponds to a WebAssembly instance, but is also used
 /// for host-defined objects.
 ///
-/// This structure is is never allocated directly but is instead managed through
+/// This structure is never allocated directly but is instead managed through
 /// an `InstanceHandle`. This structure ends with a `VMContext` which has a
 /// dynamic size corresponding to the `module` configured within. Memory
 /// management of this structure is always externalized.
@@ -410,6 +413,7 @@ impl Instance {
 
     /// Translate a module-level interned type index into an engine-level
     /// interned type index.
+    #[cfg(feature = "gc")]
     pub fn engine_type_index(&self, module_index: ModuleInternedTypeIndex) -> VMSharedTypeIndex {
         self.runtime_info.engine_type_index(module_index)
     }
@@ -424,8 +428,8 @@ impl Instance {
         unsafe { &*self.vmctx_plus_offset(self.offsets().vmctx_vmfunction_import(index)) }
     }
 
-    /// Return the index `VMTableImport`.
-    fn imported_table(&self, index: TableIndex) -> &VMTableImport {
+    /// Return the index `VMTable`.
+    fn imported_table(&self, index: TableIndex) -> &VMTable {
         unsafe { &*self.vmctx_plus_offset(self.offsets().vmctx_vmtable_import(index)) }
     }
 
@@ -439,6 +443,16 @@ impl Instance {
         unsafe { &*self.vmctx_plus_offset(self.offsets().vmctx_vmglobal_import(index)) }
     }
 
+    /// Return the indexed `VMTagImport`.
+    fn imported_tag(&self, index: TagIndex) -> &VMTagImport {
+        unsafe { &*self.vmctx_plus_offset(self.offsets().vmctx_vmtag_import(index)) }
+    }
+
+    /// Return the indexed `VMTagDefinition`.
+    fn tag_ptr(&mut self, index: DefinedTagIndex) -> NonNull<VMTagDefinition> {
+        unsafe { self.vmctx_plus_offset_mut(self.offsets().vmctx_vmtag_definition(index)) }
+    }
+
     /// Return the indexed `VMTableDefinition`.
     #[allow(dead_code)]
     fn table(&mut self, index: DefinedTableIndex) -> VMTableDefinition {
@@ -568,8 +582,8 @@ impl Instance {
 
     /// Return a pointer to the interrupts structure
     #[inline]
-    pub fn runtime_limits(&mut self) -> NonNull<Option<VmPtr<VMRuntimeLimits>>> {
-        unsafe { self.vmctx_plus_offset_mut(self.offsets().ptr.vmctx_runtime_limits()) }
+    pub fn vm_store_context(&mut self) -> NonNull<Option<VmPtr<VMStoreContext>>> {
+        unsafe { self.vmctx_plus_offset_mut(self.offsets().ptr.vmctx_store_context()) }
     }
 
     /// Return a pointer to the global epoch counter used by this instance.
@@ -578,16 +592,6 @@ impl Instance {
         unsafe { self.vmctx_plus_offset_mut(self.offsets().ptr.vmctx_epoch_ptr()) }
     }
 
-    /// Return a pointer to the GC heap base pointer.
-    pub fn gc_heap_base(&mut self) -> NonNull<Option<VmPtr<u8>>> {
-        unsafe { self.vmctx_plus_offset_mut(self.offsets().ptr.vmctx_gc_heap_base()) }
-    }
-
-    /// Return a pointer to the GC heap bound.
-    pub fn gc_heap_bound(&mut self) -> NonNull<usize> {
-        unsafe { self.vmctx_plus_offset_mut(self.offsets().ptr.vmctx_gc_heap_bound()) }
-    }
-
     /// Return a pointer to the collector-specific heap data.
     pub fn gc_heap_data(&mut self) -> NonNull<Option<VmPtr<u8>>> {
         unsafe { self.vmctx_plus_offset_mut(self.offsets().ptr.vmctx_gc_heap_data()) }
@@ -597,31 +601,33 @@ impl Instance {
         self.store = store.map(VMStoreRawPtr);
         if let Some(mut store) = store {
             let store = store.as_mut();
-            self.runtime_limits()
-                .write(Some(store.vmruntime_limits().into()));
+            self.vm_store_context()
+                .write(Some(store.vm_store_context_ptr().into()));
             #[cfg(target_has_atomic = "64")]
             self.epoch_ptr()
                 .write(Some(NonNull::from(store.engine().epoch_counter()).into()));
-            self.set_gc_heap(store.gc_store_mut().ok());
+
+            if self.env_module().needs_gc_heap {
+                self.set_gc_heap(Some(store.gc_store().expect(
+                    "if we need a GC heap, then `Instance::new_raw` should have already \
+                     allocated it for us",
+                )));
+            } else {
+                self.set_gc_heap(None);
+            }
         } else {
-            self.runtime_limits().write(None);
+            self.vm_store_context().write(None);
             #[cfg(target_has_atomic = "64")]
             self.epoch_ptr().write(None);
             self.set_gc_heap(None);
         }
     }
 
-    unsafe fn set_gc_heap(&mut self, gc_store: Option<&mut GcStore>) {
+    unsafe fn set_gc_heap(&mut self, gc_store: Option<&GcStore>) {
         if let Some(gc_store) = gc_store {
-            let heap = gc_store.gc_heap.heap_slice_mut();
-            self.gc_heap_bound().write(heap.len());
-            self.gc_heap_base()
-                .write(Some(NonNull::from(heap).cast().into()));
             self.gc_heap_data()
                 .write(Some(gc_store.gc_heap.vmctx_gc_heap_data().into()));
         } else {
-            self.gc_heap_bound().write(0);
-            self.gc_heap_base().write(None);
             self.gc_heap_data().write(None);
         }
     }
@@ -712,6 +718,17 @@ impl Instance {
         }
     }
 
+    fn get_exported_tag(&mut self, index: TagIndex) -> ExportTag {
+        ExportTag {
+            definition: if let Some(def_index) = self.env_module().defined_tag_index(index) {
+                self.tag_ptr(def_index)
+            } else {
+                self.imported_tag(index).from.as_non_null()
+            },
+            tag: self.env_module().tags[index],
+        }
+    }
+
     /// Return an iterator over the exports of this instance.
     ///
     /// Specifically, it provides access to the key-value pairs, where the keys
@@ -857,14 +874,9 @@ impl Instance {
     fn construct_func_ref(
         &mut self,
         index: FuncIndex,
-        sig: ModuleInternedTypeIndex,
+        type_index: VMSharedTypeIndex,
         into: *mut VMFuncRef,
     ) {
-        let type_index = unsafe {
-            let base = self.type_ids_array().read().as_ptr();
-            *base.add(sig.index())
-        };
-
         let func_ref = if let Some(def_index) = self.env_module().defined_func_index(index) {
             VMFuncRef {
                 array_call: self
@@ -932,7 +944,7 @@ impl Instance {
             // if we don't have to track "is-initialized" state at
             // all!
             let func = &self.env_module().functions[index];
-            let sig = func.signature;
+            let sig = func.signature.unwrap_engine_type_index();
             let func_ref = self
                 .vmctx_plus_offset_mut::<VMFuncRef>(self.offsets().vmctx_func_ref(func.func_ref));
             self.construct_func_ref(index, sig, func_ref.as_ptr());
@@ -1068,6 +1080,7 @@ impl Instance {
                             )
                         }),
                     )?,
+                    WasmHeapTopType::Cont => todo!(), // FIXME: #10248 stack switching support.
                 }
             }
         }
@@ -1409,6 +1422,14 @@ impl Instance {
             imports.globals.len(),
         );
 
+        debug_assert_eq!(imports.tags.len(), module.num_imported_tags);
+        ptr::copy_nonoverlapping(
+            imports.tags.as_ptr(),
+            self.vmctx_plus_offset_mut(offsets.vmctx_imported_tags_begin())
+                .as_ptr(),
+            imports.tags.len(),
+        );
+
         // N.B.: there is no need to initialize the funcrefs array because we
         // eagerly construct each element in it whenever asked for a reference
         // to that element. In other words, there is no state needed to track
@@ -1453,6 +1474,18 @@ impl Instance {
         for (index, _init) in module.global_initializers.iter() {
             self.global_ptr(index).write(VMGlobalDefinition::new());
         }
+
+        // Initialize the defined tags
+        let mut ptr = self.vmctx_plus_offset_mut(offsets.vmctx_tags_begin());
+        for i in 0..module.num_defined_tags() {
+            let defined_index = DefinedTagIndex::new(i);
+            let tag_index = module.tag_index(defined_index);
+            let tag = module.tags[tag_index];
+            ptr.write(VMTagDefinition::new(
+                tag.signature.unwrap_engine_type_index(),
+            ));
+            ptr = ptr.add(1);
+        }
     }
 
     fn wasm_fault(&self, addr: usize) -> Option<WasmFault> {
@@ -1507,6 +1540,11 @@ impl InstanceHandle {
         self.instance_mut().get_exported_global(export)
     }
 
+    /// Lookup a tag by index.
+    pub fn get_exported_tag(&mut self, export: TagIndex) -> ExportTag {
+        self.instance_mut().get_exported_tag(export)
+    }
+
     /// Lookup a memory by index.
     pub fn get_exported_memory(&mut self, export: MemoryIndex) -> ExportMemory {
         self.instance_mut().get_exported_memory(export)
@@ -1524,6 +1562,7 @@ impl InstanceHandle {
             EntityIndex::Global(i) => Export::Global(self.get_exported_global(i)),
             EntityIndex::Table(i) => Export::Table(self.get_exported_table(i)),
             EntityIndex::Memory(i) => Export::Memory(self.get_exported_memory(i)),
+            EntityIndex::Tag(i) => Export::Tag(self.get_exported_tag(i)),
         }
     }
 
@@ -1657,8 +1696,8 @@ impl InstanceHandle {
     ///
     /// This is provided for the original `Store` itself to configure the first
     /// self-pointer after the original `Box` has been initialized.
-    pub unsafe fn set_store(&mut self, store: NonNull<dyn VMStore>) {
-        self.instance_mut().set_store(Some(store));
+    pub unsafe fn set_store(&mut self, store: Option<NonNull<dyn VMStore>>) {
+        self.instance_mut().set_store(store);
     }
 
     /// Returns a clone of this instance.
