@@ -8,12 +8,10 @@
 //      // these fields is a compile-time constant when using `HostPtr`.
 //      magic: u32,
 //      _padding: u32, // (On 64-bit systems)
-//      runtime_limits: *const VMRuntimeLimits,
+//      vm_store_context: *const VMStoreContext,
 //      builtin_functions: *mut VMBuiltinFunctionsArray,
 //      callee: *mut VMFunctionBody,
 //      epoch_ptr: *mut AtomicU64,
-//      gc_heap_base: *mut u8,
-//      gc_heap_bound: *mut u8,
 //      gc_heap_data: *mut T, // Collector-specific pointer
 //      type_ids: *const VMSharedTypeIndex,
 //
@@ -27,16 +25,18 @@
 //      memories: [*mut VMMemoryDefinition; module.num_defined_memories],
 //      owned_memories: [VMMemoryDefinition; module.num_owned_memories],
 //      imported_functions: [VMFunctionImport; module.num_imported_functions],
-//      imported_tables: [VMTableImport; module.num_imported_tables],
+//      imported_tables: [VMTable; module.num_imported_tables],
 //      imported_globals: [VMGlobalImport; module.num_imported_globals],
+//      imported_tags: [VMTagImport; module.num_imported_tags],
 //      tables: [VMTableDefinition; module.num_defined_tables],
 //      globals: [VMGlobalDefinition; module.num_defined_globals],
+//      tags: [VMTagDefinition; module.num_defined_tags],
 //      func_refs: [VMFuncRef; module.num_escaped_funcs],
 // }
 
 use crate::{
-    DefinedGlobalIndex, DefinedMemoryIndex, DefinedTableIndex, FuncIndex, FuncRefIndex,
-    GlobalIndex, MemoryIndex, Module, OwnedMemoryIndex, TableIndex,
+    DefinedGlobalIndex, DefinedMemoryIndex, DefinedTableIndex, DefinedTagIndex, FuncIndex,
+    FuncRefIndex, GlobalIndex, MemoryIndex, Module, OwnedMemoryIndex, TableIndex, TagIndex,
 };
 use cranelift_entity::packed_option::ReservedValue;
 
@@ -69,6 +69,8 @@ pub struct VMOffsets<P> {
     pub num_imported_memories: u32,
     /// The number of imported globals in the module.
     pub num_imported_globals: u32,
+    /// The number of imported tags in the module.
+    pub num_imported_tags: u32,
     /// The number of defined tables in the module.
     pub num_defined_tables: u32,
     /// The number of defined memories in the module.
@@ -77,6 +79,8 @@ pub struct VMOffsets<P> {
     pub num_owned_memories: u32,
     /// The number of defined globals in the module.
     pub num_defined_globals: u32,
+    /// The number of defined tags in the module.
+    pub num_defined_tags: u32,
     /// The number of escaped functions in the module, the size of the func_refs
     /// array.
     pub num_escaped_funcs: u32,
@@ -86,10 +90,12 @@ pub struct VMOffsets<P> {
     imported_tables: u32,
     imported_memories: u32,
     imported_globals: u32,
+    imported_tags: u32,
     defined_tables: u32,
     defined_memories: u32,
     owned_memories: u32,
     defined_globals: u32,
+    defined_tags: u32,
     defined_func_refs: u32,
     size: u32,
 }
@@ -99,8 +105,8 @@ pub trait PtrSize {
     /// Returns the pointer size, in bytes, for the target.
     fn size(&self) -> u8;
 
-    /// The offset of the `VMContext::runtime_limits` field
-    fn vmcontext_runtime_limits(&self) -> u8 {
+    /// The offset of the `VMContext::store_context` field
+    fn vmcontext_store_context(&self) -> u8 {
         u8::try_from(align(
             u32::try_from(core::mem::size_of::<u32>()).unwrap(),
             u32::from(self.size()),
@@ -110,7 +116,7 @@ pub trait PtrSize {
 
     /// The offset of the `VMContext::builtin_functions` field
     fn vmcontext_builtin_functions(&self) -> u8 {
-        self.vmcontext_runtime_limits() + self.size()
+        self.vmcontext_store_context() + self.size()
     }
 
     /// The offset of the `array_call` field.
@@ -150,39 +156,65 @@ pub trait PtrSize {
         16
     }
 
-    // Offsets within `VMRuntimeLimits`
-
-    /// Return the offset of the `fuel_consumed` field of `VMRuntimeLimits`
+    /// Return the size of `VMTagDefinition`.
     #[inline]
-    fn vmruntime_limits_fuel_consumed(&self) -> u8 {
+    fn size_of_vmtag_definition(&self) -> u8 {
+        4
+    }
+
+    // Offsets within `VMStoreContext`
+
+    /// Return the offset of the `fuel_consumed` field of `VMStoreContext`
+    #[inline]
+    fn vmstore_context_fuel_consumed(&self) -> u8 {
         0
     }
 
-    /// Return the offset of the `epoch_deadline` field of `VMRuntimeLimits`
+    /// Return the offset of the `epoch_deadline` field of `VMStoreContext`
     #[inline]
-    fn vmruntime_limits_epoch_deadline(&self) -> u8 {
-        self.vmruntime_limits_fuel_consumed() + 8
+    fn vmstore_context_epoch_deadline(&self) -> u8 {
+        self.vmstore_context_fuel_consumed() + 8
     }
 
-    /// Return the offset of the `stack_limit` field of `VMRuntimeLimits`
+    /// Return the offset of the `stack_limit` field of `VMStoreContext`
     #[inline]
-    fn vmruntime_limits_stack_limit(&self) -> u8 {
-        self.vmruntime_limits_epoch_deadline() + 8
+    fn vmstore_context_stack_limit(&self) -> u8 {
+        self.vmstore_context_epoch_deadline() + 8
     }
 
-    /// Return the offset of the `last_wasm_exit_fp` field of `VMRuntimeLimits`.
-    fn vmruntime_limits_last_wasm_exit_fp(&self) -> u8 {
-        self.vmruntime_limits_stack_limit() + self.size()
+    /// Return the offset of the `gc_heap` field of `VMStoreContext`.
+    #[inline]
+    fn vmstore_context_gc_heap(&self) -> u8 {
+        self.vmstore_context_stack_limit() + self.size()
     }
 
-    /// Return the offset of the `last_wasm_exit_pc` field of `VMRuntimeLimits`.
-    fn vmruntime_limits_last_wasm_exit_pc(&self) -> u8 {
-        self.vmruntime_limits_last_wasm_exit_fp() + self.size()
+    /// Return the offset of the `gc_heap.base` field within a `VMStoreContext`.
+    fn vmstore_context_gc_heap_base(&self) -> u8 {
+        let offset = self.vmstore_context_gc_heap() + self.vmmemory_definition_base();
+        debug_assert!(offset < self.vmstore_context_last_wasm_exit_fp());
+        offset
     }
 
-    /// Return the offset of the `last_wasm_entry_fp` field of `VMRuntimeLimits`.
-    fn vmruntime_limits_last_wasm_entry_fp(&self) -> u8 {
-        self.vmruntime_limits_last_wasm_exit_pc() + self.size()
+    /// Return the offset of the `gc_heap.current_length` field within a `VMStoreContext`.
+    fn vmstore_context_gc_heap_current_length(&self) -> u8 {
+        let offset = self.vmstore_context_gc_heap() + self.vmmemory_definition_current_length();
+        debug_assert!(offset < self.vmstore_context_last_wasm_exit_fp());
+        offset
+    }
+
+    /// Return the offset of the `last_wasm_exit_fp` field of `VMStoreContext`.
+    fn vmstore_context_last_wasm_exit_fp(&self) -> u8 {
+        self.vmstore_context_gc_heap() + self.size_of_vmmemory_definition()
+    }
+
+    /// Return the offset of the `last_wasm_exit_pc` field of `VMStoreContext`.
+    fn vmstore_context_last_wasm_exit_pc(&self) -> u8 {
+        self.vmstore_context_last_wasm_exit_fp() + self.size()
+    }
+
+    /// Return the offset of the `last_wasm_entry_fp` field of `VMStoreContext`.
+    fn vmstore_context_last_wasm_entry_fp(&self) -> u8 {
+        self.vmstore_context_last_wasm_exit_pc() + self.size()
     }
 
     // Offsets within `VMMemoryDefinition`
@@ -231,16 +263,16 @@ pub trait PtrSize {
         0
     }
 
-    /// Return the offset to the `VMRuntimeLimits` structure
+    /// Return the offset to the `VMStoreContext` structure
     #[inline]
-    fn vmctx_runtime_limits(&self) -> u8 {
+    fn vmctx_store_context(&self) -> u8 {
         self.vmctx_magic() + self.size()
     }
 
     /// Return the offset to the `VMBuiltinFunctionsArray` structure
     #[inline]
     fn vmctx_builtin_functions(&self) -> u8 {
-        self.vmctx_runtime_limits() + self.size()
+        self.vmctx_store_context() + self.size()
     }
 
     /// Return the offset to the `callee` member in this `VMContext`.
@@ -256,25 +288,13 @@ pub trait PtrSize {
         self.vmctx_callee() + self.size()
     }
 
-    /// Return the offset to the GC heap base in this `VMContext`.
-    #[inline]
-    fn vmctx_gc_heap_base(&self) -> u8 {
-        self.vmctx_epoch_ptr() + self.size()
-    }
-
-    /// Return the offset to the GC heap bound in this `VMContext`.
-    #[inline]
-    fn vmctx_gc_heap_bound(&self) -> u8 {
-        self.vmctx_gc_heap_base() + self.size()
-    }
-
     /// Return the offset to the `*mut T` collector-specific data.
     ///
     /// This is a pointer that different collectors can use however they see
     /// fit.
     #[inline]
     fn vmctx_gc_heap_data(&self) -> u8 {
-        self.vmctx_gc_heap_bound() + self.size()
+        self.vmctx_epoch_ptr() + self.size()
     }
 
     /// The offset of the `type_ids` array pointer.
@@ -323,6 +343,8 @@ pub struct VMOffsetsFields<P> {
     pub num_imported_memories: u32,
     /// The number of imported globals in the module.
     pub num_imported_globals: u32,
+    /// The number of imported tags in the module.
+    pub num_imported_tags: u32,
     /// The number of defined tables in the module.
     pub num_defined_tables: u32,
     /// The number of defined memories in the module.
@@ -331,6 +353,8 @@ pub struct VMOffsetsFields<P> {
     pub num_owned_memories: u32,
     /// The number of defined globals in the module.
     pub num_defined_globals: u32,
+    /// The number of defined tags in the module.
+    pub num_defined_tags: u32,
     /// The number of escaped functions in the module, the size of the function
     /// references array.
     pub num_escaped_funcs: u32,
@@ -353,10 +377,12 @@ impl<P: PtrSize> VMOffsets<P> {
             num_imported_tables: cast_to_u32(module.num_imported_tables),
             num_imported_memories: cast_to_u32(module.num_imported_memories),
             num_imported_globals: cast_to_u32(module.num_imported_globals),
+            num_imported_tags: cast_to_u32(module.num_imported_tags),
             num_defined_tables: cast_to_u32(module.num_defined_tables()),
             num_defined_memories: cast_to_u32(module.num_defined_memories()),
             num_owned_memories,
             num_defined_globals: cast_to_u32(module.globals.len() - module.num_imported_globals),
+            num_defined_tags: cast_to_u32(module.tags.len() - module.num_imported_tags),
             num_escaped_funcs: cast_to_u32(module.num_escaped_funcs),
         })
     }
@@ -382,9 +408,11 @@ impl<P: PtrSize> VMOffsets<P> {
                     num_imported_tables: _,
                     num_imported_memories: _,
                     num_imported_globals: _,
+                    num_imported_tags: _,
                     num_defined_tables: _,
                     num_defined_globals: _,
                     num_defined_memories: _,
+                    num_defined_tags: _,
                     num_owned_memories: _,
                     num_escaped_funcs: _,
 
@@ -416,8 +444,10 @@ impl<P: PtrSize> VMOffsets<P> {
 
         calculate_sizes! {
             defined_func_refs: "module functions",
+            defined_tags: "defined tags",
             defined_globals: "defined globals",
             defined_tables: "defined tables",
+            imported_tags: "imported tags",
             imported_globals: "imported globals",
             imported_tables: "imported tables",
             imported_functions: "imported functions",
@@ -436,19 +466,23 @@ impl<P: PtrSize> From<VMOffsetsFields<P>> for VMOffsets<P> {
             num_imported_tables: fields.num_imported_tables,
             num_imported_memories: fields.num_imported_memories,
             num_imported_globals: fields.num_imported_globals,
+            num_imported_tags: fields.num_imported_tags,
             num_defined_tables: fields.num_defined_tables,
             num_defined_memories: fields.num_defined_memories,
             num_owned_memories: fields.num_owned_memories,
             num_defined_globals: fields.num_defined_globals,
+            num_defined_tags: fields.num_defined_tags,
             num_escaped_funcs: fields.num_escaped_funcs,
             imported_functions: 0,
             imported_tables: 0,
             imported_memories: 0,
             imported_globals: 0,
+            imported_tags: 0,
             defined_tables: 0,
             defined_memories: 0,
             owned_memories: 0,
             defined_globals: 0,
+            defined_tags: 0,
             defined_func_refs: 0,
             size: 0,
         };
@@ -492,14 +526,18 @@ impl<P: PtrSize> From<VMOffsetsFields<P>> for VMOffsets<P> {
             size(imported_functions)
                 = cmul(ret.num_imported_functions, ret.size_of_vmfunction_import()),
             size(imported_tables)
-                = cmul(ret.num_imported_tables, ret.size_of_vmtable_import()),
+                = cmul(ret.num_imported_tables, ret.size_of_vmtable()),
             size(imported_globals)
                 = cmul(ret.num_imported_globals, ret.size_of_vmglobal_import()),
+            size(imported_tags)
+                = cmul(ret.num_imported_tags, ret.size_of_vmtag_import()),
             size(defined_tables)
                 = cmul(ret.num_defined_tables, ret.size_of_vmtable_definition()),
             align(16),
             size(defined_globals)
                 = cmul(ret.num_defined_globals, ret.ptr.size_of_vmglobal_definition()),
+            size(defined_tags)
+                = cmul(ret.num_defined_tags, ret.ptr.size_of_vmtag_definition()),
             size(defined_func_refs) = cmul(
                 ret.num_escaped_funcs,
                 ret.ptr.size_of_vm_func_ref(),
@@ -546,23 +584,23 @@ impl<P: PtrSize> VMOffsets<P> {
     }
 }
 
-/// Offsets for `VMTableImport`.
+/// Offsets for `VMTable`.
 impl<P: PtrSize> VMOffsets<P> {
     /// The offset of the `from` field.
     #[inline]
-    pub fn vmtable_import_from(&self) -> u8 {
+    pub fn vmtable_from(&self) -> u8 {
         0 * self.pointer_size()
     }
 
     /// The offset of the `vmctx` field.
     #[inline]
-    pub fn vmtable_import_vmctx(&self) -> u8 {
+    pub fn vmtable_vmctx(&self) -> u8 {
         1 * self.pointer_size()
     }
 
-    /// Return the size of `VMTableImport`.
+    /// Return the size of `VMTable`.
     #[inline]
-    pub fn size_of_vmtable_import(&self) -> u8 {
+    pub fn size_of_vmtable(&self) -> u8 {
         2 * self.pointer_size()
     }
 }
@@ -638,6 +676,21 @@ impl<P: PtrSize> VMOffsets<P> {
     }
 }
 
+/// Offsets for `VMTagImport`.
+impl<P: PtrSize> VMOffsets<P> {
+    /// The offset of the `from` field.
+    #[inline]
+    pub fn vmtag_import_from(&self) -> u8 {
+        0 * self.pointer_size()
+    }
+
+    /// Return the size of `VMTagImport`.
+    #[inline]
+    pub fn size_of_vmtag_import(&self) -> u8 {
+        1 * self.pointer_size()
+    }
+}
+
 /// Offsets for `VMContext`.
 impl<P: PtrSize> VMOffsets<P> {
     /// The offset of the `tables` array.
@@ -664,6 +717,12 @@ impl<P: PtrSize> VMOffsets<P> {
         self.imported_globals
     }
 
+    /// The offset of the `tags` array.
+    #[inline]
+    pub fn vmctx_imported_tags_begin(&self) -> u32 {
+        self.imported_tags
+    }
+
     /// The offset of the `tables` array.
     #[inline]
     pub fn vmctx_tables_begin(&self) -> u32 {
@@ -688,6 +747,12 @@ impl<P: PtrSize> VMOffsets<P> {
         self.defined_globals
     }
 
+    /// The offset of the `tags` array.
+    #[inline]
+    pub fn vmctx_tags_begin(&self) -> u32 {
+        self.defined_tags
+    }
+
     /// The offset of the `func_refs` array.
     #[inline]
     pub fn vmctx_func_refs_begin(&self) -> u32 {
@@ -708,12 +773,11 @@ impl<P: PtrSize> VMOffsets<P> {
             + index.as_u32() * u32::from(self.size_of_vmfunction_import())
     }
 
-    /// Return the offset to `VMTableImport` index `index`.
+    /// Return the offset to `VMTable` index `index`.
     #[inline]
     pub fn vmctx_vmtable_import(&self, index: TableIndex) -> u32 {
         assert!(index.as_u32() < self.num_imported_tables);
-        self.vmctx_imported_tables_begin()
-            + index.as_u32() * u32::from(self.size_of_vmtable_import())
+        self.vmctx_imported_tables_begin() + index.as_u32() * u32::from(self.size_of_vmtable())
     }
 
     /// Return the offset to `VMMemoryImport` index `index`.
@@ -732,6 +796,13 @@ impl<P: PtrSize> VMOffsets<P> {
             + index.as_u32() * u32::from(self.size_of_vmglobal_import())
     }
 
+    /// Return the offset to `VMTagImport` index `index`.
+    #[inline]
+    pub fn vmctx_vmtag_import(&self, index: TagIndex) -> u32 {
+        assert!(index.as_u32() < self.num_imported_tags);
+        self.vmctx_imported_tags_begin() + index.as_u32() * u32::from(self.size_of_vmtag_import())
+    }
+
     /// Return the offset to `VMTableDefinition` index `index`.
     #[inline]
     pub fn vmctx_vmtable_definition(&self, index: DefinedTableIndex) -> u32 {
@@ -763,6 +834,13 @@ impl<P: PtrSize> VMOffsets<P> {
             + index.as_u32() * u32::from(self.ptr.size_of_vmglobal_definition())
     }
 
+    /// Return the offset to the `VMTagDefinition` index `index`.
+    #[inline]
+    pub fn vmctx_vmtag_definition(&self, index: DefinedTagIndex) -> u32 {
+        assert!(index.as_u32() < self.num_defined_tags);
+        self.vmctx_tags_begin() + index.as_u32() * u32::from(self.ptr.size_of_vmtag_definition())
+    }
+
     /// Return the offset to the `VMFuncRef` for the given function
     /// index (either imported or defined).
     #[inline]
@@ -790,10 +868,11 @@ impl<P: PtrSize> VMOffsets<P> {
         self.vmctx_vmfunction_import(index) + u32::from(self.vmfunction_import_vmctx())
     }
 
-    /// Return the offset to the `from` field in `VMTableImport` index `index`.
+    /// Return the offset to the `from` field in the imported `VMTable` at index
+    /// `index`.
     #[inline]
-    pub fn vmctx_vmtable_import_from(&self, index: TableIndex) -> u32 {
-        self.vmctx_vmtable_import(index) + u32::from(self.vmtable_import_from())
+    pub fn vmctx_vmtable_from(&self, index: TableIndex) -> u32 {
+        self.vmctx_vmtable_import(index) + u32::from(self.vmtable_from())
     }
 
     /// Return the offset to the `base` field in `VMTableDefinition` index `index`.
@@ -838,6 +917,12 @@ impl<P: PtrSize> VMOffsets<P> {
     pub fn vmctx_vmglobal_import_from(&self, index: GlobalIndex) -> u32 {
         self.vmctx_vmglobal_import(index) + u32::from(self.vmglobal_import_from())
     }
+
+    /// Return the offset to the `from` field in `VMTagImport` index `index`.
+    #[inline]
+    pub fn vmctx_vmtag_import_from(&self, index: TagIndex) -> u32 {
+        self.vmctx_vmtag_import(index) + u32::from(self.vmtag_import_from())
+    }
 }
 
 /// Offsets for `VMDrcHeader`.
